//
// purpose: See original comments
//  author: KOL version,  2004, Thaddy de Koning
// Remarks: This translation is kindly authorised by Milenko Mitrovic;
//

    (*********************************************************************
     *  dspCompressor.pas                                                *
     *                                                                   *
     *  This unit is Part of the DC-DSP Component Pack v1.0              *
     *                                                                   *
     *  author    : Milenko Mitrovic                                     *
     *  email     : dcoder@dsp-worx.de                                   *
     *  web       : http://dsp-worx.de                                   *
     *  date      : 24-07-2003                                           *
     *                                                                   *
     *  The contents of this file are used with permission, subject to   *
     *  the Mozilla Public License Version 1.1 (the "License"); you may  *
     *  not use this file except in compliance with the License. You may *
     *  obtain a copy of the License at                                  *
     *  http://www.mozilla.org/MPL/MPL-1.1.html                          *
     *                                                                   *
     *  Software distributed under the License is distributed on an      *
     *  "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or   *
     *  implied. See the License for the specific language governing     *
     *  rights and limitations under the License.                        *
     *                                                                   *
     *  based on the C++ Audacity Compressor Cpde.                       *
     *  (C) The Audacity Project under the Terms of the GPL !!!          *
     *  http://audacity.sourceforge.net/                                 *
     *                                                                   *
     *  Delphi Conversion:                                               *
     *  (C) 2003, 2004 Milenko Mitrovic <dcoder@dsp-worx.de>             *
     *                                                                   *
     *********************************************************************)
{
  @abstract(DSP Filter to Compress (Amplify) Audio Data.)
  @author(Milenko Mitrovic <dcoder@dsp-worx.de>)
  @created(Sep 01, 2002)
  @lastmod(Apr 02, 2004)
}

unit koldspCompressor;

interface

uses
  kol,koldspConst, koldspUtils, kolMath, Windows;

type
  { TDCCompressor - DSP Compressor Filter. }
  PDCCompressor = ^TDCCompressor;
  TDCCompressor = object(Tobj)
  protected
    {@exclude}
    fAttackTime : Single;
    {@exclude}
    fDecayTime : Single;
    {@exclude}
    fThresholdDB : Single;
    {@exclude}
    fRatio : Single;
    {@exclude}
    fGainDB : Single;
    {@exclude}
    fCircleSize : integer;
    {@exclude}
    fCirclePos : integer;
    {@exclude}
    fRMSSum : Double;
    {@exclude}
    fMult : Double;
    {@exclude}
    fThreshold : Double;
    {@exclude}
    fGain : Double;
    {@exclude}
    fInvRatio : Double;
    {@exclude}
    fEnabled : Boolean;
    {@exclude}
    fProcessMessages : Boolean;
    {@exclude}
    fSampleSize : Cardinal;
    {@exclude}
    fFloor: Double;
    {@exclude}
    fCircle: PDoubleArray;
    {@exclude}
    fLevelCircle: PDoubleArray;
    {@exclude}
    fLastLevel: Double;
    {@exclude}
    fAttackFactor: Double;
    {@exclude}
    fDecayFactor: Double;
    {@exclude}
    fFollow: PDoubleArray;
    {@exclude}
    fFollowSize: integer;
    {@exclude}
    fBuffer: PDoubleArray;
    {@exclude}
    function DoCompression(value: Single; env: Double): single;
    {@exclude}
    procedure Follow(x: Single; outEnv: PDoubleArray; maxBack: integer);
    {@exclude}
    function AvgCircle(value: Single): Single;
    {@exclude}
    procedure DoDSP(Buffer : Pointer; Size : Integer; Samplerate : integer; Bits : Byte; Channels : Byte; Float : Boolean);
    {@exclude}
    procedure InitVars(Samplerate : integer; Channels: integer);
    {@exclude}
    procedure SetAttackTime(aAttack: Single);
  public
    { Destroys the Instance. }
    destructor Destroy; virtual;
    { Call this Method to Process Audio Data. }
    procedure Process(Buffer : Pointer; Size : Integer; Samplerate : integer;Bits : Byte; Channels : Byte; Float : Boolean);
    {@exclude}
    function get_Enabled(out aEnabled: BOOL): HRESULT; stdcall;
    {@exclude}
    function set_Enabled(aEnabled: BOOL): HRESULT; stdcall;
    {@exclude}
    function get_AttackTime(out aAttackTime: Single): HRESULT; stdcall;
    {@exclude}
    function set_AttackTime(aAttackTime: Single): HRESULT; stdcall;
    {@exclude}
    function get_DecayTime(out aDecayTime: Single): HRESULT; stdcall;
    {@exclude}
    function set_DecayTime(aDecayTime: Single): HRESULT; stdcall;
    {@exclude}
    function get_ThresholdDB(out aThresholdDB: Single): HRESULT; stdcall;
    {@exclude}
    function set_ThresholdDB(aThresholdDB: Single): HRESULT; stdcall;
    {@exclude}
    function get_Ratio(out aRatio: Single): HRESULT; stdcall;
    {@exclude}
    function set_Ratio(aRatio: Single): HRESULT; stdcall;
    {@exclude}
    function get_GainDB(out aGainDB: Single): HRESULT; stdcall;
    {@exclude}
    function set_GainDB(aGainDB: Single): HRESULT; stdcall;
    { Enables or Disables the Filter. }
    property Enabled : Boolean read fEnabled write fEnabled;
    { On large Buffersize Processing the Main Application wouldnt respond until
      the DSP is done. by using "ProcessMessages" and "SampleSize" the DSP will
      be splitted into N(N = SampleSize) Number of Samples and then Processed.
      After every Process if ProcessMessages is Enabled Application.Processmessages
      will be called. }
    property ProcessMessages : Boolean read fProcessMessages write fProcessMessages;
    { Sets the SampleSize so that the Buffer will be splitted to Process an amount
      of Data only on every Process. Set SampleSize to 0 to disable it. Prevent
      using small size Values. Use at least 1024 or more. }
    property SampleSize : Cardinal read fSampleSize write fSampleSize;
    { Specifys the Attack Time in MilliSeconds. }
    property AttackTime : Single read fAttackTime write SetAttackTime;
    { Specifys the Release Time in Milli Seconds. }
    property DecayTime : Single read fDecayTime write fDecayTime;
    { Specifys the Threshold in Dezibel. }
    property ThresholdDB : Single read fThresholdDB write fThresholdDB;
    { Specifys the Amplification Ratio. }
    property Ratio : Single read fRatio write fRatio;
    { Specifys the maximum Amplification in Dezibel. }
    property GainDB : Single read fGainDB write fGainDB;
  end;

  function NewDCCompressor(AOwner: PControl):PDCCompressor;

implementation



function NewDCCompressor(AOwner: PControl):PDCCompressor;
var
  i: integer;
begin
  New(Result,Create);
  AOwner.Add2autofree(Result);
  with Result^ do
  begin
    fCircleSize := 100;
    fFollowSize := 2;
    fFollow := AllocMem(fFollowSize);
    fBuffer := AllocMem(fFollowSize);
    fCircle := AllocMem(fCircleSize * SizeOf(Double));
    fLevelCircle := AllocMem(fCircleSize * SizeOf(Double));

    for i := 0 to fCircleSize -1 do
    begin
      fCircle[i] := 0.0;
      fLevelCircle[i] := fFloor;
    end;

    fFloor := 0.001;
    fAttackTime := 0.2;
    fDecayTime := 1.0;
    fThresholdDB := -12.0;
    fRatio := 2.0;
    fGainDB := 0.0;

    fCirclePos := 0;
    fRMSSum := 0.0;
    fLastLevel := 0.0;

    fEnabled := False;
    fProcessMessages := False;
    fSampleSize := DefaultSampleSize;
  end;
end;

destructor TDCCompressor.Destroy;
begin
  FreeMemory(fLevelCircle);
  FreeMemory(fCircle);
  FreeMemory(fBuffer);
  FreeMemory(fFollow);
  inherited Destroy;
end;

procedure TDCCompressor.InitVars(Samplerate : integer; Channels: integer);
begin
  fThreshold := power(10.0, fThresholdDB/10); // factor of 10 because it's power
  fGain := power(10.0, fGainDB/20); // factor of 20 because it's amplitude
  fAttackFactor := exp(-ln(fFloor) / (Samplerate * Channels * fAttackTime + 0.5));
  fDecayFactor := exp(ln(fFloor) / (Samplerate * Channels * fDecayTime + 0.5));
end;

procedure TDCCompressor.Process(Buffer : Pointer; Size : Integer; Samplerate : integer; Bits : Byte; Channels : Byte; Float : Boolean);
var
  SplitBuffer : PChar;
  SizeLeft : integer;
  SplitSize : integer;
  CurrentSize : integer;
begin
  if not fEnabled then Exit;
  InitVars(SampleRate, Channels);
  if fSampleSize = 0 then
  begin
    DoDSP(Buffer,Size,SampleRate,Bits,Channels,Float);
  end else
  begin
    SplitBuffer := Buffer;
    SplitSize := fSampleSize * (Bits div 8) * Channels;
    SizeLeft := Size;
    while SizeLeft > 0 do
    begin
      if SizeLeft > SplitSize then CurrentSize := SplitSize
                              else CurrentSize := SizeLeft;
      DoDSP(@SplitBuffer[Size - SizeLeft],CurrentSize,Samplerate,Bits,Channels,Float);
      if fProcessMessages then Applet.ProcessMessages;
      dec(SizeLeft,SplitSize);
    end;
  end;
end;

function TDCCompressor.AvgCircle(value: Single): Single;
var
  level: Single;
begin
  fRMSSum := fRMSSum - fCircle[fCirclePos];
  fCircle[fCirclePos] := value*value;
  fRMSSum := fRMSSum + fCircle[fCirclePos];
  level := sqrt(fRMSSum/fCircleSize);
  fLevelCircle[fCirclePos] := level;
  fCirclePos := (fCirclePos+1) mod fCircleSize;

  Result := level;
end;

procedure TDCCompressor.Follow(x: Single; outEnv: PDoubleArray; maxBack: integer);
var
  level: Single;
  high: Single;
  low: Single;
  attackInverse: Double;
  temp: Double;
  backtrack: integer;
  ptr: PDoubleArray;
  t: integer;
  i: integer;
  ok: Boolean;
begin
  ok := false;
  level := AvgCircle(x);
  high := fLastLevel * fAttackFactor;
  low := fLastLevel * fDecayFactor;
  if (low < fFloor) then low := fFloor;
  if (level < low) then outEnv^[0] := low
  else if (level < high) then outEnv^[0] := level
  else
  begin
    attackInverse := 1.0 / fAttackFactor;
    temp := level * attackInverse;

    backtrack := 50;
    if (backtrack > maxBack) then backtrack := maxBack;

    t := -1;
    ptr := @outEnv[t];

    for i := 0 to backtrack-3 do
    begin
      if (ptr[0] < temp) then
      begin
        ptr := @ptr[t];
        ptr[0] := temp;
        temp := temp * attackInverse;
      end else
      begin
        ok := true;
        break;
      end;
    end;

    if ((not ok) and (backtrack>1) and (ptr[0] < temp)) then
    begin
      temp := ptr[0];
      for i := 0 to backtrack-2 do
      begin
        ptr := @ptr[1];
        temp := temp * fAttackFactor;
        ptr[0] := temp;
       end;
    end else
    begin
      outEnv[0] := level;
    end;
  end;
  fLastLevel := outEnv[0];
end;

function TDCCompressor.DoCompression(value: Single; env: Double): single;
begin
  if (env > fThreshold)
    then Result := Clip_32F(value * (fGain * power(fThreshold/env, 1.0/fRatio)))
    else Result := Clip_32F(value * fGain);
end;

procedure TDCCompressor.DoDSP(Buffer : Pointer; Size : Integer; Samplerate : integer; Bits : Byte; Channels : Byte; Float : Boolean);
var
  Buf8  : PByteArray;
  Buf16 : PSmallIntArray;
  Buf24 : PInteger24Array;
  Buf32i : PIntegerArray;
  Buf32 : PFloatArray;
  NumSamples : integer;
  i : integer;
  preSeed: integer;
begin
  Buf8 := PByteArray(Buffer);
  Buf16 := PSmallIntArray(Buffer);
  Buf24 := PInteger24Array(Buffer);
  Buf32 := PFloatArray(Buffer);
  Buf32i := PIntegerArray(Buffer);

  NumSamples := Size div (Bits div 8);// div Channels;
  if fFollowSize < (NumSamples * SizeOf(Double)) then
  begin
    fFollowSize := NumSamples * SizeOf(Double);
    fBuffer := ReallocMemory(fBuffer,fFollowSize);
    fFollow := ReallocMemory(fFollow,fFollowSize);
  end;

  for i := 0 to NumSamples -1 do
  begin
    case Bits of
      8:  fBuffer[i] := (Buf8[i] - 128) / 128;
      16: fBuffer[i] := Buf16[i] / 32768;
      24: fBuffer[i] := Cvt24BitTo32(Buf24[i]) / 8388608;
      32: if Float then fBuffer[i] := Buf32[i];
                   else fBuffer[i] := Buf32i[i] / 2147483648;
    end;
  end;

  if (fLastLevel = 0.0) then
  begin
    preSeed := fCircleSize;
    if (preSeed > NumSamples) then preSeed := NumSamples;
    for i := 0 to preSeed -1 do AvgCircle(fBuffer[i]);
  end;

  for i := 0 to NumSamples -1 do
    Follow(fBuffer[i], @fFollow[i], i);

  for i := 0 to NumSamples -1 do
     fBuffer[i] := DoCompression(fBuffer[i],fFollow[i]);

  for i := 0 to NumSamples -1 do
  begin
    case Bits of
      8:  Buf8[i] := Clip_8(Round(fBuffer[i] * 128)) + 128;
      16: Buf16[i] := Clip_16(Round(fBuffer[i] * 32768));
      24: Buf24[i] := Cvt32BitTo24(Clip_24(Round(fBuffer[i] * 8388608)));
      32: if Float then Buf32[i] := fBuffer[i]
                   else Buf32i[i] := Clip_32(Int64(Round(fBuffer[i] * 2147483648)));
    end;
  end;
end;

procedure TDCCompressor.SetAttackTime(aAttack: Single);
begin
  fAttackTime := aAttack;
  fRMSSum := 1.0;
end;
(*** IDCCompressor ************************************************************)
function TDCCompressor.get_Enabled(out aEnabled: BOOL): HRESULT; stdcall;
begin
  Result := S_OK;
  aEnabled := fEnabled;
end;

function TDCCompressor.set_Enabled(aEnabled: BOOL): HRESULT; stdcall;
begin
  Result := S_OK;
  fEnabled := aEnabled;
end;

function TDCCompressor.get_AttackTime(out aAttackTime: Single): HRESULT; stdcall;
begin
  Result := S_OK;
  aAttackTime := fAttackTime;
end;

function TDCCompressor.set_AttackTime(aAttackTime: Single): HRESULT; stdcall;
begin
  Result := S_OK;
  SetAttackTime(aAttackTime);
end;

function TDCCompressor.get_DecayTime(out aDecayTime: Single): HRESULT; stdcall;
begin
  Result := S_OK;
  aDecayTime := fDecayTime;
end;

function TDCCompressor.set_DecayTime(aDecayTime: Single): HRESULT; stdcall;
begin
  Result := S_OK;
  fDecayTime := aDecayTime;
end;

function TDCCompressor.get_ThresholdDB(out aThresholdDB: Single): HRESULT; stdcall;
begin
  Result := S_OK;
  aThresholdDB := fThresholdDB;
end;

function TDCCompressor.set_ThresholdDB(aThresholdDB: Single): HRESULT; stdcall;
begin
  Result := S_OK;
  fThresholdDB := aThresholdDB;
end;

function TDCCompressor.get_Ratio(out aRatio: Single): HRESULT; stdcall;
begin
  Result := S_OK;
  aRatio := fRatio;
end;

function TDCCompressor.set_Ratio(aRatio: Single): HRESULT; stdcall;
begin
  Result := S_OK;
  fRatio := aRatio;
end;

function TDCCompressor.get_GainDB(out aGainDB: Single): HRESULT; stdcall;
begin
  Result := S_OK;
  aGainDB := fGainDB;
end;

function TDCCompressor.set_GainDB(aGainDB: Single): HRESULT; stdcall;
begin
  Result := S_OK;
  fGainDB := aGainDB;
end;

end.

